package dev;

import dev.util.ServerValidator;

import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.nio.charset.StandardCharsets;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

public final class Server implements Runnable {

    private static final int CLIENTS_MAX = 3;
    private static final String PW_DEFAULT = "geheim";
    private static final int PORT_DEFAULT = 50_000;

    private static final int TIMEOUT_IN_MS = 30_000;
    static final long UPDATE_INTERVAL_IN_MS = 200L;
    private static final long DELAY_IN_MS = 225L;

    private static Server instance = null;

    static boolean attemptShutdown(String pw) {
        if (instance == null) return false;

        if (instance.isValidPw(pw)) {
            instance.sessionPool.shutdown();
            if (Server.instance.sessionPool.getActiveCount() == 0) {
                Server.instance.sessionPool.shutdownNow();
            }
            return true;
        } else {
            return false;
        }
    }

    static void updateHeartbeat(long currentTimeMillis) {
        Server.instance.latestHeartbeat = Math.max(currentTimeMillis, Server.instance.latestHeartbeat);
    }

    private boolean isValidPw(String pw) {
        return pw.equals(getPw());
    }

    static String currentTime() {
        return SimpleDateFormat.getTimeInstance().format(new Date());
    }

    private ThreadPoolExecutor sessionPool;
    private long latestHeartbeat;
    private final String pw;
    private final int port;

    private ServerSocket welcomeSocket;

    private int getPort() {
        return port;
    }

    private String getPw() {
        return pw;
    }

    Server() {
        this(PORT_DEFAULT, PW_DEFAULT);
        System.out.println("* Starte Server mit Standard-Parametern *");
    }

    Server(int port, String pw) {
        this.port = port;
        this.pw = pw;
        Server.instance = this;
    }

    private void denyConnection(String reason) {
        try {
            denyConnection(welcomeSocket.accept(), reason);
        } catch (IOException ex) { /* _ */ }
    }

    private void denyConnection(Socket socket, String reason) {
        if (socket == null) {
            denyConnection(reason); // öhh, das ruft eine Methode auf, die wiederum diese Methode aufruft.. gefährlich
            return;
        }

        try (final DataOutputStream outputStream = new DataOutputStream(socket.getOutputStream())) {
            System.out.printf("[%s] Sitzung verweigert für: %s\n", currentTime(), socket.getRemoteSocketAddress());

            byte[] msgBytes = String.format("Verbindung abgelehnt: %s", reason).getBytes(StandardCharsets.UTF_8);
            for (byte b : msgBytes) {
                if (b == '\n' || b == '\r') break;
                outputStream.write(b);
            }
            outputStream.write('\n');

            outputStream.flush();
        } catch (IOException ex) { /* _ */ }
    }

    private void startServer() throws IOException {
        this.welcomeSocket = new ServerSocket();
        welcomeSocket.setReceiveBufferSize(ServerValidator.BYTES_MAX);
        welcomeSocket.bind(new InetSocketAddress(getPort()), 1);
        this.sessionPool = new ThreadPoolExecutor(CLIENTS_MAX, CLIENTS_MAX, 0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<>(CLIENTS_MAX));

        System.out.printf("Server gestartet. Port: %d, PW: %s\n", getPort(), getPw());
        System.out.printf("Erlaubte aktive Verbindungen: %d; Erlaubte Warteschlangengröße: %d\n",
                sessionPool.getCorePoolSize(),
                sessionPool.getMaximumPoolSize() - sessionPool.getCorePoolSize()
        );
    }

    @Override
    public void run() {
        Thread.currentThread().setName(this.toString());

        try {
            startServer();

            Thread welcomeThread = new Thread(() -> {
                while (true) {
                    if (sessionPool.getActiveCount() >= CLIENTS_MAX || sessionPool.isTerminating()) {
                        denyConnection("Keine freien Plätze, später versuchen");
                    }

                    if (sessionPool.getActiveCount() < CLIENTS_MAX && !sessionPool.isTerminating()) {
                        Socket socket = null;
                        try {
                            Thread.sleep(DELAY_IN_MS);
                            socket = new Socket();
                            socket.setReceiveBufferSize(ServerValidator.BYTES_MAX);
                            socket.setSendBufferSize(ServerValidator.BYTES_MAX);
                            socket = welcomeSocket.accept();
                            Session session = new Session(socket);
                            sessionPool.execute(session);
                        } catch (SocketException ignore) {
                            denyConnection(socket, "Kann keine Verbindung herstellen");
                        } catch (RejectedExecutionException ignore) {
                            denyConnection(socket, "keine freien Plätze");
                        } catch (IOException e) {
                            e.printStackTrace();
                        } catch (InterruptedException ignore) {
                            return;
                        }
                    }
                }
            });

            welcomeThread.start();

            while (true) {
                if (sessionPool.isTerminated()) {
                    if (welcomeThread.isAlive()) {
                        welcomeThread.interrupt();
                    }

                    System.out.println("Keine aktiven Clients mehr, Server wird sofort heruntergefahren...");
                    sessionPool.shutdownNow();
                    System.exit(0);
                }

                if (sessionPool.isTerminating() || sessionPool.isShutdown()) {
                    if (welcomeThread.isAlive() || !welcomeThread.isInterrupted()) {
                        System.out.println("Schließe Anmelde-Socket. Es werden keine neuen Verbindungen zugelassen.");
                        welcomeThread.interrupt();
                    }

                    System.out.println("Warte auf Timeout der verbundenen Clients...");
                    boolean terminated = false;

                    while (!terminated) {
                        try {
                            terminated = sessionPool.awaitTermination(UPDATE_INTERVAL_IN_MS, TimeUnit.MILLISECONDS);
                            if ((System.currentTimeMillis() - latestHeartbeat) > TIMEOUT_IN_MS) {
                                if (welcomeThread.isAlive()) welcomeThread.interrupt();

                                System.out.println("Timeout, Beenden wird erzwungen");
                                sessionPool.shutdownNow();
                                System.exit(0);
                            }

                        } catch (InterruptedException e) { /* _ */ }
                    }

                    System.out.println("Keine aktiven Sitzungen mehr, Server wird jetzt heruntergefahren...");
                    sessionPool.shutdownNow();
                    System.exit(0);

                }

                try {
                    Thread.sleep(UPDATE_INTERVAL_IN_MS);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }

            }
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    @Override
    public String toString() {
        return String.format("Server @ Port %d", getPort());
    }
}
