package dev;

import dev.util.ServerExceptions;
import dev.util.ServerValidator;
import dev.util.StringHelper;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

final class Session implements Runnable {

    private final Socket socket;
    private DataInputStream inputStream;
    private DataOutputStream outputStream;
    private boolean logoutTriggered;

    private String threadName;

    private void setThreadName(String threadName) {
        this.threadName = threadName;
    }

    private String getThreadName() {
        return this.threadName;
    }

    Session(Socket socket) {
        this.socket = socket;
        this.logoutTriggered = false;
    }

    private void endSession() {
        logoutTriggered = true;
    }

    private void shutdownServer(String pw) {
        if (Server.attemptShutdown(pw)) {
            try {
                sendMsg("Server wird bald heruntergefahren");
            } catch (IOException e) { /* _ */ }
        } else {
            try {
                sendMsg("Falsches Passwort: Verbindung wird getrennt, Server läuft weiter");
            } catch (IOException e) { /* _ */ }
        }
        endSession();
    }

    // Wäre vielleicht besser in einem Builder oder im ServerValidator aufgehoben
    Map<Command, String> parseCmdFromString(String text) throws ServerExceptions.SyntaxError, ServerExceptions.UnknownCommand {

        final Map<Command, String> cmd = new HashMap<>(1);
        String[] input = text.split(Command.DELIMITER, 2);

        Command cmdEnum;
        try {
            cmdEnum = Command.valueOf(input[0]);
        } catch (IllegalArgumentException e) {
            throw new ServerExceptions.UnknownCommand();
        }

        // if-else würde hier reichen, es gibt hier im Grunde nur 2 Fälle
        switch (cmdEnum) {
            case BYE:
                if (input.length != 1) {
                    throw new ServerExceptions.SyntaxError("BYE nur ohne Parameter erlaubt");
                }
                cmd.put(cmdEnum, null);
                break;
            case LOWERCASE:
                if (input.length != 2 || input[1].isEmpty()) {
                    throw new ServerExceptions.SyntaxError("LOWERCASE: Text benötigt");
                }
            case REVERSE:
                if (input.length != 2 || input[1].isEmpty()) {
                    throw new ServerExceptions.SyntaxError("REVERSE: Text benötigt");
                }
            case SHUTDOWN:
                if (input.length != 2 || input[1].isEmpty()) {
                    throw new ServerExceptions.SyntaxError("SHUTDOWN: Passwort benötigt");
                }
            case UPPERCASE:
                if (input.length != 2 || input[1].isEmpty()) {
                    throw new ServerExceptions.SyntaxError("UPPERCASE: Text benötigt");
                }
            default:
                if (!ServerValidator.PARAM_VALID(cmdEnum, input[1])) {
                    throw new ServerExceptions.SyntaxError(String.format(
                            "Ungültiger Parameter, %s benötigt Text und erlaubt maximal %d Bytes",
                            cmdEnum.name(), ServerValidator.maxParamBytesInCmd(cmdEnum))
                    );
                }

                cmd.put(cmdEnum, input[1]);
        }
        return cmd;
    }

    private String processCmd(Map<Command, String> cmd)
            throws ServerExceptions.SyntaxError, ServerExceptions.UnknownCommand {
        for (Map.Entry<Command, String> cmdEntry : cmd.entrySet()) {
            switch (cmdEntry.getKey()) {
                case BYE:
                    endSession();
                    return "OK BYE";
                case SHUTDOWN:
                    shutdownServer(cmdEntry.getValue());
                    return "OK BYE";
                case LOWERCASE:
                    return "OK " + StringHelper.toLowerCase(cmdEntry.getValue());
                case UPPERCASE:
                    return "OK " + StringHelper.toUpperCase(cmdEntry.getValue());
                case REVERSE:
                    return "OK " + StringHelper.reverseString(cmdEntry.getValue());
                default:
                    break;
            }
        }
        throw new ServerExceptions.UnknownCommand();
    }

    private void processRequest(String request) throws IOException {
        try {
            Map<Command, String> parsedRequest = parseCmdFromString(request);
            sendMsg(processCmd(parsedRequest));
        } catch (ServerExceptions.UnknownCommand | ServerExceptions.SyntaxError e) {
            sendMsg(e.getMessage());
        }
    }

    private void greetClient() throws IOException {
        sendMsg(Command.getCmdInfos());
    }

    private void sendMsg(final String msg) throws IOException {
        byte[] msgBytes = msg.getBytes(StandardCharsets.UTF_8);

        for (byte b : msgBytes) {
            outputStream.write(b);
        }
        outputStream.write('\n');
        outputStream.flush();
    }

    private String receiveMsg() throws IOException, ServerExceptions.SyntaxError {
        byte[] inputBuffer = new byte[ServerValidator.BYTES_MAX];

        int offset;
        for (offset = 0; offset < ServerValidator.BYTES_MAX; offset++) {
            byte b = inputStream.readByte();
            inputBuffer[offset] = b;

            if (b == '\n') break;

            if (offset >= ServerValidator.BYTES_MAX - 1) {
                System.out.printf(Locale.GERMANY,
                        "[%s] Nachricht zu lang (%s @ %s)\n",
                        Server.currentTime(),
                        getThreadName(),
                        socket.getRemoteSocketAddress()
                );
                throw new ServerExceptions.SyntaxError("Nachricht zu lang oder endet nicht mit \\n!");
            }
        }

        return new String(inputBuffer, 0, offset, StandardCharsets.UTF_8);
    }

    private void initSession(String threadName) throws IOException {
        setThreadName(threadName);

        System.out.printf(Locale.GERMANY,
                "[%s] Sitzung gestartet (%s @ %s)\n",
                Server.currentTime(),
                getThreadName(),
                socket.getRemoteSocketAddress()
        );

        inputStream = new DataInputStream(socket.getInputStream());
        outputStream = new DataOutputStream(socket.getOutputStream());

        greetClient();
    }

    @Override
    public void run() {
        try {
            initSession(Thread.currentThread().getName());

            Thread clientMessages = new Thread(() -> {
                try {
                    while (!Thread.currentThread().isInterrupted()) {
                        String msg = receiveMsg();
                        System.out.printf("[%s] %s: %s\n", Server.currentTime(), Thread.currentThread().getName(), msg);
                        Server.updateHeartbeat(System.currentTimeMillis());
                        processRequest(msg);
                    }
                } catch (IOException e) {
                    System.out.printf("[%s] %s: Verbindung getrennt\n", Server.currentTime(), getThreadName());
                    endSession();
                } catch (ServerExceptions.SyntaxError ex) {
                    try {
                        sendMsg(ex.getMessage());
                        sendMsg("Sitzung geschlossen. Neu verbinden");
                    } catch (IOException couldNotSendMsg) { /* _ */ } finally {
                        endSession();
                    }
                }
            });

            clientMessages.start();

            while (true) {
                if (logoutTriggered) {
                    if (clientMessages.isAlive()) {
                        clientMessages.interrupt();
                    }
                    if (!socket.isClosed()) {
                        socket.shutdownOutput();
                        socket.shutdownInput();
                        socket.close();
                    }
                    return;
                }

                try {
                    Thread.sleep(Server.UPDATE_INTERVAL_IN_MS);
                } catch (InterruptedException e) {
                    return;
                }
            }
        } catch (IOException e) {
            System.err.println("Kann Sitzung nicht initialisieren");
            e.printStackTrace();
        }
    }
}
