package dev;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.net.Socket;
import java.util.HashMap;
import java.util.Map;

import static dev.util.ServerExceptions.SyntaxError;
import static dev.util.ServerExceptions.UnknownCommand;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

class SessionTest {
    private Session session;
    private Map<Command, String> command;

    @BeforeEach
    void setUp() {
        this.session = new Session(new Socket());
        this.command = new HashMap<>();
    }

    @Test
    void denyUnknownCommands() {
        assertThrows(UnknownCommand.class, () -> session.parseCmdFromString("enum existiert nicht"));
    }

    @Test
    void denyWrongCase() {
        assertThrows(UnknownCommand.class, () -> session.parseCmdFromString("bye"));
    }

    @Test
    void byeAccept() throws Exception {
        command.put(Command.BYE, null);
        assertEquals(command, session.parseCmdFromString("BYE"));
    }

    @Test
    void byeDenyParams() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("BYE" + Command.DELIMITER + "text"));
    }

    @Test
    void shutdownAccept() throws Exception {
        command.put(Command.SHUTDOWN, "passwort");
        assertEquals(command, session.parseCmdFromString("SHUTDOWN" + Command.DELIMITER + "passwort"));
    }

    @Test
    void shutdownAcceptDelimiter() throws Exception {
        String msg = "pwteil1" + Command.DELIMITER + "pwteil2" + Command.DELIMITER;
        command.put(Command.SHUTDOWN, msg);
        assertEquals(command, session.parseCmdFromString("SHUTDOWN" + Command.DELIMITER + msg));
    }

    @Test
    void shutdownDeny0Param() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("SHUTDOWN"));
    }

    @Test
    void shutdownDenyEmptyParam() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("SHUTDOWN" + Command.DELIMITER + ""));
    }

    @Test
    void lowercaseAccept() throws Exception {
        command.put(Command.LOWERCASE, "text");
        assertEquals(command, session.parseCmdFromString("LOWERCASE" + Command.DELIMITER + "text"));
    }

    @Test
    void lowercaseAcceptDelimiter() throws Exception {
        String msg = "tEiL1" + Command.DELIMITER + "TeIl2" + Command.DELIMITER;
        command.put(Command.LOWERCASE, msg);
        assertEquals(command, session.parseCmdFromString("LOWERCASE" + Command.DELIMITER + msg));
    }

    @Test
    void lowercaseDeny0Param() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("LOWERCASE"));
    }

    @Test
    void lowercaseDenyEmptyParam() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("LOWERCASE" + Command.DELIMITER + ""));
    }

    @Test
    void uppercaseAccept() throws Exception {
        command.put(Command.UPPERCASE, "text");
        assertEquals(command, session.parseCmdFromString("UPPERCASE" + Command.DELIMITER + "text"));
    }

    @Test
    void uppercaseAcceptDelimiter() throws Exception {
        String msg = "tEiL1" + Command.DELIMITER + "TeIl2" + Command.DELIMITER;
        command.put(Command.UPPERCASE, msg);
        assertEquals(command, session.parseCmdFromString("UPPERCASE" + Command.DELIMITER + msg));
    }

    @Test
    void uppercaseDeny0Params() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("UPPERCASE"));
    }

    @Test
    void uppercaseDenyEmptyParam() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("UPPERCASE" + Command.DELIMITER + ""));
    }

    @Test
    void reverseAccept() throws Exception {
        command.put(Command.REVERSE, "text");
        assertEquals(command, session.parseCmdFromString("REVERSE" + Command.DELIMITER + "text"));
    }

    @Test
    void reverseAcceptDelimiter() throws Exception {
        String msg = "tEiL1" + Command.DELIMITER + "TeIl2" + Command.DELIMITER;
        command.put(Command.REVERSE, msg);
        assertEquals(command, session.parseCmdFromString("REVERSE" + Command.DELIMITER + msg));
    }

    @Test
    void reverseDeny0ParamNot() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("REVERSE"));
    }

    @Test
    void reverseDenyEmptyParam() {
        assertThrows(SyntaxError.class, () -> session.parseCmdFromString("REVERSE" + Command.DELIMITER + ""));
    }

}