package dev.util;

import org.junit.jupiter.api.Test;

import static dev.Command.*;
import static dev.util.ServerValidator.*;
import static dev.util.StringHelper.generateSingleByteString;
import static org.junit.jupiter.api.Assertions.*;

class ServerValidatorTest {
    private static final int DELIMITER_BYTES = StringHelper.countBytesInString(DELIMITER);

    @Test
    void portAcceptValidPort()  { assertTrue(isPortValid("50000")); }

    @Test
    void portAcceptLeading0()   { assertTrue(isPortValid("050000")); }

    @Test
    void portDenyNotANumber()   { assertFalse(isPortValid("KEINPORT")); }

    @Test
    void portDenyTooSmall()     { assertFalse(isPortValid(String.valueOf(ServerValidator.PORT_MIN - 1))); }

    @Test
    void portDenyTooBig()       { assertFalse(isPortValid(String.valueOf(ServerValidator.PORT_MAX + 1))); }

    @Test
    void pwAcceptValidPw()      { assertTrue(isPwValid("passwort")); }

    @Test
    void pwAcceptSpecialChars() { assertTrue(isPwValid("_'@€§$%&/()=?\\\"µäöüß!-.é`´?")); }

    @Test
    void pwDenyEmpty()          { assertFalse(isPwValid("")); }

    @Test
    void pwDenyNewline()        { assertFalse(isPwValid("asd\nsd")); }

    /* Ich denke dei allen Aufrufen von generateSingleByteString('<CHAR>', <ANZAHL>) wäre es besser:
        - "<Zeichen>".repeat(<ANZAHL>) von String::repeat zu benutzen
        - den Wert einmalig als langen static final String fest zu legen, und dann Substring der nötigen Länge zu holen
        - Wert aus Random mit fixem Seed zu generieren, falls auch mehr Eingaben getestet werden sollen
     */

    @Test
    void pwAcceptMaxLength() {
        assertTrue(isPwValid(generateSingleByteString('a', maxParamBytesInCmd(SHUTDOWN))));
    }

    @Test
    void pwDenyTooLong() {
        assertFalse(isPwValid(generateSingleByteString('A', maxParamBytesInCmd(SHUTDOWN) + 1)));
    }

    /* Ich denke, alle folgenden maxBytes-Tests könnten parametrisiert/dynamisch statt per Hand getestet werden,
    da die Werte von festen Regeln abhängen:
        - ob Parameter erlaubt sind (wenn nicht, dann maxBytes = 0)
        - wie lang der Befehl + Delimiter ist
        - wie viele Bytes insgesamt pro Befehl zulässig sind (hier 255)
     */

    @Test
    void maxBytesBye() {
        assertEquals(0, maxParamBytesInCmd(BYE));
    }

    @Test
    void maxBytesByeDenyParams() {
        assertFalse(PARAM_VALID(BYE, "test"));
    }

    @Test
    void maxBytesShutdown() {
        assertEquals(BYTES_MAX - "SHUTDOWN".length() - DELIMITER_BYTES, maxParamBytesInCmd(SHUTDOWN));
    }

    @Test
    void maxBytesShutdownAcceptMaxLength() {
        assertTrue(PARAM_VALID(SHUTDOWN, generateSingleByteString('a', maxParamBytesInCmd(SHUTDOWN))));
    }

    @Test
    void maxBytesShutdownDenyTooLong() {
        assertFalse(PARAM_VALID(SHUTDOWN, generateSingleByteString('a', maxParamBytesInCmd(SHUTDOWN) + 1)));
    }

    @Test
    void maxBytesLowercase() {
        assertEquals(BYTES_MAX - "LOWERCASE".length() - DELIMITER_BYTES, maxParamBytesInCmd(LOWERCASE));
    }

    @Test
    void maxBytesLowercaseAcceptMaxLength() {
        assertTrue(PARAM_VALID(LOWERCASE, generateSingleByteString('a', maxParamBytesInCmd(LOWERCASE))));
    }

    @Test
    void maxBytesLowercaseDenyTooLong() {
        assertFalse(PARAM_VALID(LOWERCASE, generateSingleByteString('a', maxParamBytesInCmd(LOWERCASE) + 1)));
    }

    @Test
    void maxBytesUppercase() {
        assertEquals(BYTES_MAX - "UPPERCASE".length() - DELIMITER_BYTES, maxParamBytesInCmd(UPPERCASE));
    }

    @Test
    void maxBytesUppercaseAcceptMaxLength() {
        assertTrue(PARAM_VALID(UPPERCASE, generateSingleByteString('a', maxParamBytesInCmd(UPPERCASE))));
    }

    @Test
    void maxBytesUppercaseDenyTooLong() {
        assertFalse(PARAM_VALID(UPPERCASE, generateSingleByteString('a', maxParamBytesInCmd(UPPERCASE) + 1)));
    }

    @Test
    void maxBytesReverse() {
        assertEquals(BYTES_MAX - "REVERSE".length() - DELIMITER_BYTES, maxParamBytesInCmd(REVERSE));
    }

    @Test
    void maxBytesReverseAcceptMaxLength() {
        assertTrue(PARAM_VALID(REVERSE, generateSingleByteString('a', maxParamBytesInCmd(REVERSE))));
    }

    @Test
    void maxBytesReverseDenyTooLong() {
        assertFalse(PARAM_VALID(REVERSE, generateSingleByteString('a', maxParamBytesInCmd(REVERSE) + 1)));
    }

}